﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Sokoban
{
    abstract class Heuristic<S, P> where S : State<S, P> where P : Problem<S, P>
    {
        public abstract void Init(P problem);
        public abstract int GetValue(S state);
    }

    class MaxHeuristic<S, P> : Heuristic<S, P> where S : State<S, P> where P : Problem<S, P>
    {
        List<Heuristic<S, P>> heuristics;

        public MaxHeuristic(List<Heuristic<S, P>> heuristics)
        {
            this.heuristics = heuristics;
        }

        public override int GetValue(S state)
        {
            return heuristics.Max(h => h.GetValue(state));
        }

        public override void Init(P problem)
        {
            foreach (var item in heuristics)
            {
                item.Init(problem);
            }
        }
    }

    class WeightedHeuristic<S, P> : Heuristic<S, P> where S : State<S, P> where P : Problem<S, P>
    {
        private readonly Heuristic<S, P> h;
        private readonly int weight;

        public WeightedHeuristic(Heuristic<S, P> h, int weight)
        {
            this.h = h;
            this.weight = weight;
        }

        public override void Init(P problem)
        {
            h.Init(problem);
        }

        public override int GetValue(S state)
        {
            return weight * h.GetValue(state);
        }
    }

    abstract class DistanceHeuristic : Heuristic<SokobanState, SokobanProblem>
    {
        private int limit;
        protected Dictionary<Coords, Dictionary<Coords, int>> distancesSokoban;
        protected Dictionary<Coords, Dictionary<Coords, int>> distancesCrate;
        protected void computeDistances(Board b)
        {
            b.FindForbidenTiles();
            distancesSokoban = new Dictionary<Coords, Dictionary<Coords, int>>();
            distancesCrate = new Dictionary<Coords, Dictionary<Coords, int>>();
            limit = b.size.x * b.size.y * 2;
            for (int i = 0; i < b.size.x; i++)
            {
                for (int j = 0; j < b.size.y; j++)
                {
                    if (b[i, j].cellState == CellState.obstacle)
                        continue;
                    UpdateDistance(i, j, i, j, 0, true);
                    UpdateDistance(i, j, i, j, 0, false);

                    Coords target = new Coords(), origin = new Coords();
                    foreach (var dir in SokobanProblem.allDirections)
                    {
                        target.x = (byte)i;
                        target.y = (byte)j;
                        target.Change(dir);
                        if (b[target].cellState == CellState.obstacle)
                            continue;
                        UpdateDistance(i, j, target.x, target.y, 1, true);

                        origin.x = target.x;
                        origin.y = target.y;
                        origin.Change(dir);
                        if (b[origin].cellState == CellState.obstacle)
                            continue;
                        UpdateDistance(target.x, target.y, i, j, 1, false);
                    }
                    /*
                    if (b[i + 1, j].state != CellState.obstacle)
                    {
                        UpdateDistance(i, j, i + 1, j, 1, true);
                        UpdateDistance(i, j, i + 1, j, 1, false);
                    }
                    if (b[i, j + 1].state != CellState.obstacle)
                    {
                        UpdateDistance(i, j, i, j + 1, 1, true);
                        UpdateDistance(i, j, i, j + 1, 1, false);
                    }
                      */
                }
            }

            for (int x1 = 0; x1 < b.size.x; x1++)
                for (int y1 = 0; y1 < b.size.y; y1++)
                    for (int x2 = 0; x2 < b.size.x; x2++)
                        for (int y2 = 0; y2 < b.size.y; y2++)
                            for (int x3 = 0; x3 < b.size.x; x3++)
                                for (int y3 = 0; y3 < b.size.y; y3++)
                                {
                                    int dist_2_1 = getDistance(x2, y2, x1, y1, true);
                                    int dist_1_3 = getDistance(x1, y1, x3, y3, true);
                                    int dist_2_3 = getDistance(x2, y2, x3, y3, true);
                                    if (dist_2_3 > dist_2_1 + dist_1_3)
                                        UpdateDistance(x2, y2, x3, y3, dist_2_1 + dist_1_3, true);

                                    //if (Solver.isCorner(new Coords(x1, y1), b))
                                    if (!b[x1, y1].IsAccesibleToCrate)
                                        continue;
                                    dist_2_1 = getDistance(x2, y2, x1, y1, false);
                                    dist_1_3 = getDistance(x1, y1, x3, y3, false);
                                    dist_2_3 = getDistance(x2, y2, x3, y3, false);
                                    if (dist_2_3 > dist_2_1 + dist_1_3)
                                        UpdateDistance(x2, y2, x3, y3, dist_2_1 + dist_1_3, false);
                                }
        }

        private int getDistance(int x1, int y1, int x2, int y2, bool isForSokoban)
        {
            Coords c1 = new Coords(x1, y1),
            c2 = new Coords(x2, y2);
            if (isForSokoban)
            {
                if (!distancesSokoban.ContainsKey(c1) || !distancesSokoban[c1].ContainsKey(c2))
                    return limit;
                return distancesSokoban[c1][c2];
            }
            else
            {
                if (!distancesCrate.ContainsKey(c1) || !distancesCrate[c1].ContainsKey(c2))
                    return limit;
                return distancesCrate[c1][c2];
            }
        }

        /// <summary>
        /// Updates distance from [x1,y1] to [x2,y2]. Distances for sokoban are symmetric, for crates they are not.
        /// </summary>
        /// <param name="x1"></param>
        /// <param name="y1"></param>
        /// <param name="x2"></param>
        /// <param name="y2"></param>
        /// <param name="distance"></param>
        /// <param name="isForSokoban"></param>
        private void UpdateDistance(int x1, int y1, int x2, int y2, int distance, bool isForSokoban)
        {
            Coords c1 = new Coords(x1, y1),
            c2 = new Coords(x2, y2);
            if (isForSokoban)
            {
                if (!distancesSokoban.ContainsKey(c1))
                    distancesSokoban.Add(c1, new Dictionary<Coords, int>());

                if (!distancesSokoban.ContainsKey(c2))
                    distancesSokoban.Add(c2, new Dictionary<Coords, int>());

                if (!distancesSokoban[c1].ContainsKey(c2))
                    distancesSokoban[c1].Add(c2, distance);
                else distancesSokoban[c1][c2] = distance;

                if (!distancesSokoban[c2].ContainsKey(c1))
                    distancesSokoban[c2].Add(c1, distance);
                else distancesSokoban[c2][c1] = distance;
            }
            else
            {
                if (!distancesCrate.ContainsKey(c1))
                    distancesCrate.Add(c1, new Dictionary<Coords, int>());

                /*if (!distancesCrate.ContainsKey(c2))
                    distancesCrate.Add(c2, new Dictionary<Coords, int>());
                 //distance for crates are NOT symmetric!
                */
                if (!distancesCrate[c1].ContainsKey(c2))
                    distancesCrate[c1].Add(c2, distance);
                else distancesCrate[c1][c2] = distance;
                /*
                if (!distancesCrate[c2].ContainsKey(c1))
                    distancesCrate[c2].Add(c1, distance);
                else distancesCrate[c2][c1] = distance;
                 */
            }
        }

        protected int getDistance(Coords c1, Coords c2, bool isForSokoban)
        {
            if (isForSokoban)
            {
                if (!distancesSokoban.ContainsKey(c1) || !distancesSokoban[c1].ContainsKey(c2))
                    return limit;
                return distancesSokoban[c1][c2];
            }
            else
            {
                if (!distancesCrate.ContainsKey(c1) || !distancesCrate[c1].ContainsKey(c2))
                    return limit;
                return distancesCrate[c1][c2];
            }
        }
    }

    class distanceTester : DistanceHeuristic
    {
        public override void Init(SokobanProblem s)
        {
            computeDistances(s.Board);
        }

        public override int GetValue(SokobanState state)
        {
            return 0;
        }

        public int GetDistance(int i, int j, int k, int l, bool isForSokoban)
        {
            return getDistance(new Coords(i, j), new Coords(k, l), isForSokoban);
        }
    }

    class SimpleDistanceHeuristic : DistanceHeuristic
    {
        Coords[] targets;
        public override void Init(SokobanProblem s)
        {
            var b = s.Board;
            computeDistances(b);
            List<Coords> targetsList = new List<Coords>();
            for (int i = 0; i < b.size.x; i++)
            {
                for (int j = 0; j < b.size.y; j++)
                {
                    if (b[i, j].cellState == CellState.target)
                        targetsList.Add(new Coords(i, j));
                }
            }
            targets = targetsList.ToArray();
        }

        public override int GetValue(SokobanState sokState)
        {
            int toNearestCrate = int.MaxValue;
            for (int i = 0; i < sokState.Crates.Length; i++)
            {
                int dist = getDistance(sokState.Crates[i], sokState.SokobanPos, true);
                if (toNearestCrate > dist)
                    toNearestCrate = dist;
            }
            int result = toNearestCrate;
            foreach (var item in targets)
            {
                toNearestCrate = int.MaxValue;
                for (int i = 0; i < sokState.Crates.Length; i++)
                {
                    int dist = getDistance(sokState.Crates[i], item, true);
                    if (toNearestCrate > dist)
                        toNearestCrate = dist;
                }
                result += toNearestCrate;
            }
            return result;
        }
    }

    class AdvancedDistanceHeuristic : DistanceHeuristic
    {
        Coords[] targets;
        Board b;
        public override void Init(SokobanProblem p)
        {
            this.b = p.Board;
            computeDistances(b);
            List<Coords> targetsList = new List<Coords>();
            for (int i = 0; i < b.size.x; i++)
            {
                for (int j = 0; j < b.size.y; j++)
                {
                    if (b[i, j].cellState == CellState.target)
                        targetsList.Add(new Coords(i, j));
                }
            }
            targets = targetsList.ToArray();
        }

        public override int GetValue(SokobanState state)
        {
            List<int> notProcessedCrates = new List<int>();
            int toNearestCrate = int.MaxValue;
            for (int i = 0; i < state.Crates.Length; i++)
            {
                notProcessedCrates.Add(i);
                if (b[state.Crates[i]].cellState == CellState.target)
                    continue;
                int dist = getDistance(state.Crates[i], state.SokobanPos, true);
                if (toNearestCrate > dist)
                    toNearestCrate = dist;
            }
            if (toNearestCrate == int.MaxValue) //is goal state
                return 0;
            int result = toNearestCrate;
            foreach (var item in targets)
            {
                toNearestCrate = int.MaxValue;
                int nearestCrate = -1;
                for (int i = 0; i < notProcessedCrates.Count; i++)
                {
                    int dist = getDistance(state.Crates[notProcessedCrates[i]], item, false);
                    if (toNearestCrate > dist)
                    {
                        toNearestCrate = dist;
                        nearestCrate = i;
                    }
                }
                result += toNearestCrate;
                notProcessedCrates.RemoveAt(nearestCrate);
            }
            return result;
        }
    }

    class BlindHeuristic<S, P> : Heuristic<S, P> where P : Problem<S, P> where S : State<S, P>
    {
        public override void Init(P problem)
        {

        }

        public override int GetValue(S state)
        {
            return 0;
        }
    }

    class MyHeuristic : Heuristic<SokobanState, SokobanProblem>
    {
        public override int GetValue(SokobanState state)
        {
            throw new NotImplementedException();
        }

        public override void Init(SokobanProblem problem)
        {
            throw new NotImplementedException();
        }
    }
}
