﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Sokoban
{
    class StateCostPair<S, P> where P : Problem<S, P> where S : State<S, P>
    {
        public S State;
        public int TransitionCost;

        public override int GetHashCode()
        {
            return State.GetHashCode() + TransitionCost;
        }

        public StateCostPair(S s, int cost)
        {
            this.State = s;
            this.TransitionCost = cost;
        }
    }

    abstract class State<S, P> where P : Problem<S, P> where S : State<S, P>
    {
        public P problem;

		public abstract bool IsGoalState();

        public abstract void GetSuccessors(List<StateCostPair<S, P>> result);

        public abstract S Clone();
    }

    abstract class Problem<S, P> where P : Problem<S, P> where S : State<S, P>
	{
        public S InitialState;
    }

	class SokobanProblem : Problem<SokobanState, SokobanProblem>
	{
		public bool useSymmetryBreaking, useAdvancedSolving, useDeadEndDetection;
		public AdvancedSuccessorGenerator advancedSuccessorGenerator = new AdvancedSuccessorGenerator();
		public Board Board;
		public static List<Direction> allDirections = new Direction[] { Direction.down, Direction.up, Direction.left, Direction.right }.ToList();
		public static List<List<Direction>> allNeighbourDirections =
			new List<Direction>[] { new Direction[] { Direction.down, Direction.left}.ToList(),
				new Direction[] { Direction.left, Direction.up}.ToList(),
				new Direction[] { Direction.up, Direction.right}.ToList(),
				new Direction[] { Direction.right, Direction.down}.ToList()}.ToList();

		public SokobanProblem(bool useSymmetryBreaking, bool useAdvancedSolving, bool useDeadEndDetection, Board board)
		{
			this.useSymmetryBreaking = useSymmetryBreaking;
			this.useAdvancedSolving = useAdvancedSolving;
			this.useDeadEndDetection = useDeadEndDetection;
			Board = board;
			InitialState = new SokobanState(Board.Crates.Select(c => c.Location).ToArray(), Board.sokobanPos, this);
		}

		public class AdvancedSuccessorGenerator
		{
			private Queue<Coords> openList;
			private Queue<int> g_values;
			private HashSet<Coords> closedList;

            public AdvancedSuccessorGenerator()
            {
				this.closedList = new HashSet<Coords>();
				this.g_values = new Queue<int>();
				this.openList = new Queue<Coords>();
            }

			public IEnumerable<(Coords sokobanPos, Direction intendedMoveDirection, int actionCost)> GetAdvanceActions(SokobanState state)
			{
				closedList.Clear();
				openList.Clear();
				openList.Enqueue(state.SokobanPos);
				g_values.Enqueue(0);
				while (openList.Count > 0)
				{
					Coords current = openList.Dequeue();
					int currentGVal = g_values.Dequeue();

					if (closedList.Contains(current))
						continue;
					closedList.Add(current);

					foreach (var dir in allDirections)
					{
						Coords target = current.GetNeighbour(dir);
						if (state.problem.Board[target].cellState == CellState.obstacle)
							continue;
						bool isCrate = false;
						if (state.Crates.Any(c => c.Equals(target)))//is crate
						{
							isCrate = true;
							yield return (current, dir, currentGVal + 1);
						}

						if (!isCrate)
						{
							if (!closedList.Contains(target))
							{
								openList.Enqueue(target);
								g_values.Enqueue(currentGVal + 1);
							}
						}
					}
				}
			}
		}

		/// <summary>
		/// Each action is represented as a triple (current position of Sokoban, intended direction, action cost). 
		/// This method filters out actions from <paramref name="actions"/> that are not applicable to <paramref name="state"/> and for the rest it creates the 
		/// corresponding successors states together with transition costs and returns these via <paramref name="resultPlaceholder"/>.
		/// </summary>
		/// <param name=""></param>
		/// <param name="resultPlaceholder"></param>
		public void ApplyActions(SokobanState state, IEnumerable<(Coords sokobanPos, Direction intendedMoveDirection, int actionCost)> actions,
			List<StateCostPair<SokobanState, SokobanProblem>> resultPlaceholder)
		{
			resultPlaceholder.Clear();
			foreach (var action in actions)
			{
				if (Board[action.sokobanPos, action.intendedMoveDirection].cellState == CellState.obstacle)
					continue;   //cannot move - the target cell is blocked

				Coords sokobanTarget = action.sokobanPos.GetNeighbour(action.intendedMoveDirection);
				var crateIndex = state.CrateIndex(sokobanTarget);
				if (crateIndex >= 0) //is push action
				{
					Coords pushTarget = sokobanTarget.GetNeighbour(action.intendedMoveDirection);

					if (Board[pushTarget].cellState == CellState.obstacle || !Board[pushTarget].IsAccesibleToCrate ||
						state.Crates.Any(c => c.Equals(pushTarget)))
						continue;   //cannot push - the target is blocked
					else
					{
						var res = state.Clone();
						res.SokobanPos = sokobanTarget;
						res.Crates[crateIndex] = pushTarget;

						if (useSymmetryBreaking)
							sortState(res.Crates);
						resultPlaceholder.Add(new StateCostPair<SokobanState, SokobanProblem>(res, action.actionCost));
					}
				}
				else // is move action
				{
					var res = state.Clone();
					res.SokobanPos = sokobanTarget;
					resultPlaceholder.Add(new StateCostPair<SokobanState, SokobanProblem>(res, 1));
				}
			}
		}

		protected void sortState(Coords[] crates)
		{
			Array.Sort(crates, 0, crates.Length, Coords.GetComparer());
		}

		public List<Direction> ExtractSolution(List<SokobanState> goalPath)
        {
			if (useAdvancedSolving)
				return extractSolutionAdvanced(goalPath);
			return extractSolution(goalPath);
        }

		public static Direction GetOpositeDirection(Direction dir)
		{
			switch (dir)
			{
				case Direction.down:
					return Direction.up;
				case Direction.left:
					return Direction.right;
				case Direction.right:
					return Direction.left;
				case Direction.up:
					return Direction.down;
			}
			throw new ArgumentException();
		}

		public static Direction GetTransferDirection(Coords from, Coords to)
		{
			if (to.x == from.x)
			{
				if (to.y < from.y)
					return Direction.up;
				else return Direction.down;
			}
			else
			{
				if (to.x < from.x)
					return Direction.left;
				else return Direction.right;
			}
		}

		protected Direction getTransferAction(SokobanState successor, SokobanState predecessor)
		{
			return GetTransferDirection(predecessor.SokobanPos, successor.SokobanPos);
		}

		protected List<Direction> findPath(Coords from, Coords to, SokobanState state)
		{
			List<Direction> result = new List<Direction>();
			Dictionary<Coords, Direction> predecessoronPath = new Dictionary<Coords, Direction>();
			List<Coords> pathSearchFrontier = new List<Coords>();

			pathSearchFrontier.Add(from);
			while (pathSearchFrontier.Count > 0)
			{
				Coords c = pathSearchFrontier[0];
				pathSearchFrontier.RemoveAt(0);

				if (c.Equals(to))
					break;

				foreach (var dir in allDirections)
				{
					Coords succ = c.GetNeighbour(dir);
					if (state.problem.Board[succ].cellState == CellState.obstacle || state.Crates.Any(c => c.Equals(succ)) || predecessoronPath.ContainsKey(succ))
						continue;
					predecessoronPath.Add(succ, dir);
					pathSearchFrontier.Add(succ);
				}
			}
			Coords current = to;
			while (!current.Equals(from))
			{
				result.Insert(0, predecessoronPath[current]);
				current.Change(GetOpositeDirection(predecessoronPath[current]));
			}

			return result;
		}

		protected List<Direction> extractSolutionAdvanced(List<SokobanState> states)
		{
			List<Direction> solution = new List<Direction>();

			SokobanState currentState = states.Last();
            for (int i = states.Count - 2; i >= 0; i--)
            {
				SokobanState pred = states[i];

				Direction d = GetTransferDirection(currentState.Crates.ToList()
					.Where(c => !pred.Crates.Any(q => q.Equals(c))).Single(), currentState.SokobanPos);
				Coords c = currentState.SokobanPos.GetNeighbour(d);
				solution.Insert(0, GetOpositeDirection(d));
				solution.InsertRange(0, findPath(pred.SokobanPos, c, pred));
				currentState = pred;
			}

			return solution;
		}

		protected List<Direction> extractSolution(List<SokobanState> states)
		{
			List<Direction> solution = new List<Direction>();
			SokobanState currentState = states.Last();
			for (int i = states.Count - 2; i >= 0; i--)
			{
				SokobanState pred = states[i];
				solution.Insert(0, getTransferAction(currentState, pred));
				currentState = pred;

			}
			return solution;
		}

	}

	class SokobanState : State<SokobanState, SokobanProblem>
    {
        public Coords[] Crates;
		public Coords SokobanPos;

        public override bool Equals(object obj)
        {
            if (obj is SokobanState s)
            {
				return this.SokobanPos.Equals(s.SokobanPos) && Coords.GetListComparer().Equals(this.Crates, s.Crates);
            }
			return false;
        }

        public override int GetHashCode()
        {
            unchecked
			{
				return Coords.GetListComparer().GetHashCode(this.Crates) + this.SokobanPos.GetHashCode() * 31;
			}
        }

        public Board Input => this.problem.Board;
		
        public override SokobanState Clone()
        {
			var copy = new Coords[this.Crates.Length];
			Parallel.For(0, this.Crates.Length, i => copy[i] = new Coords(this.Crates[i].x, this.Crates[i].y));

			return new SokobanState(copy, new Coords(this.SokobanPos.x, this.SokobanPos.y), problem);
		}

		public override void GetSuccessors(List<StateCostPair<SokobanState, SokobanProblem>> result)
        {
			if (problem.useAdvancedSolving)
				problem.ApplyActions(this, problem.advancedSuccessorGenerator.GetAdvanceActions(this).ToList(), result);
			else
				problem.ApplyActions(this, getAtions(), result);
        }

        public SokobanState(Coords[] Crates, Coords SokobanPos, SokobanProblem p)
        {
            this.Crates = Crates;
			this.SokobanPos = SokobanPos;
			this.problem = p;
		}

		protected IEnumerable<(Coords sokobanPos, Direction intendedMoveDirection, int actionCost)> getAtions()
		{
			foreach (var direction in SokobanProblem.allDirections)
			{
				yield return (this.SokobanPos, direction, 1);
			}
		}

		public override bool IsGoalState()
		{
			return Crates.All(c => problem.Board[c].cellState == CellState.target);
		}

        public override string ToString()
        {
			return string.Join(" ", Crates.Select(c => c.ToString())) + " | " + SokobanPos;
        }

		/// <summary>
		/// Returns index of a crate in "Crates" whose location is <paramref name="c"/>. If no such crate exists it returns -1;
		/// </summary>
		/// <param name="c"></param>
		/// <returns></returns>
		public int CrateIndex(Coords c)
        {
            for (int i = 0; i < Crates.Length; i++)
            {
				if (Crates[i].Equals(c))
					return i;
            }
			return -1;
        }
    }
}
