﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices.ComTypes;
using System.Text;

namespace Sokoban
{
	abstract class Solver<S, P> where P : Problem<S, P> where S : State<S, P>
    {
		public abstract List<S> Solve(P problem, TimeSpan maxTime);

		public bool quiet = false;

		protected void printMsg(string msg)
        {
			if (!quiet)
                Console.WriteLine(msg);
        }

	}

	abstract class HeuristicSolver<S, P> : Solver<S, P> where P : Problem<S, P> where S : State<S, P>
	{
		protected Heuristic<S, P> h;

		public void SetHeuristic(Heuristic<S, P> h)
		{
			this.h = h;
		}

	}

	class Astar<S, P> : HeuristicSolver<S, P> where P : Problem<S, P> where S : State<S, P>
	{
        protected IHeap<int, S> openList;
		protected List<StateCostPair<S, P>> successorPlaceHolder;
		protected S initialState;
		protected Dictionary<S, StateValPair<S, P>> visited;
		protected P problem;

		protected TimeSpan timelimit;

		S searchResult;
		long expansions, minimalHValue, maxGValue;
		protected System.Diagnostics.Stopwatch stopWatch;
		double totalExpansions, totalBranching,	heuristicCalls, totalHeuristicValue;

		public Astar()
        {
			successorPlaceHolder = new List<StateCostPair<S, P>>();
        }


        public override List<S> Solve(P problem, TimeSpan maxtime)
        {
            this.init(problem);
            Solve(maxtime);
            return extractSolution(searchResult);
        }

        protected void init(P problem)
        {
			this.visited = new Dictionary<S, StateValPair<S, P>>();
			this.initialState = problem.InitialState;
			this.problem = problem;

            h.Init(problem);

            //openList = new RegularBinaryHeap<S>();
			openList = new RedBlackTreeHeap<S>();
		}

		protected virtual void AddToOpenList(S state, int gVal, int hVal)
        {
			if (hVal < int.MaxValue)
				openList.Insert(gVal + hVal, state);
		}

		/// <summary>
		/// Adds the node to closed list. Returns false if the same node with the same or less GVal was already added. Returns true otherwise.
		/// </summary>
		/// <param name="state"></param>
		/// <param name="predecessor"></param>
		/// <param name="gVal"></param>
		/// <returns></returns>
		protected virtual bool AddToClosedList(S state, S predecessor, int gVal)
		{
			if (visited.ContainsKey(state))
			{
				var node = visited[state];
				if (node.gValue <= gVal)
					return false;
				else
				{
					//a shorter path found to previously visited node. Updating its GVal and predecessor.
					node.gValue = gVal;
					node.predecessor = predecessor;
					visited[state] = node;
				}
			}
			else
				visited.Add(state, new StateValPair<S, P>(predecessor, gVal));
			return true;
		}

		private void printProgress()
        {
			printMsg("\n");
			printMsg("Time elapsed: " + stopWatch.Elapsed.TotalSeconds + " seconds");
			printMsg("Closed nodes: " + visited.Keys.Count);
			printMsg("Open nodes: " + openList.Size());
			printMsg("Max GValue: " + maxGValue);
			printMsg("Speed: " + visited.Count / stopWatch.Elapsed.TotalSeconds + " nodes per second");
			printMsg("Average branching: " + totalBranching / totalExpansions);
			printMsg("Average heuristic value: " + totalHeuristicValue / heuristicCalls);
			printMsg("Minimal heuristic value: " + minimalHValue);
			//printMsg("Hash code calls: " + comparer.hashCodeCalls);
			//printMsg("Equals calls: " + comparer.equalsCalls + " (" + ((double)comparer.hashCodeCalls / comparer.equalsCalls) + " : 1)");
			//printMsg("Hash colissions: " + comparer.colissions +  " (" + ((double)visited.Keys.Count / comparer.colissions) + " : 1)");
		}

		private void printResult(S goalState)
		{
			printMsg("");
			printMsg("Search ended in " + stopWatch.Elapsed.TotalSeconds + " seconds");
			printMsg("Solution length " + visited[goalState].gValue);
			printMsg("Closed nodes: " + visited.Keys.Count);
			printMsg("Open nodes: " + openList.Size());
			printMsg("Max GValue: " + maxGValue);
			printMsg("Speed: " + visited.Count / stopWatch.Elapsed.TotalSeconds + " nodes per second");
			printMsg("Average branching: " + totalBranching / totalExpansions);
			printMsg("Average heuristic value: " + totalHeuristicValue / heuristicCalls);
			printMsg("Minimal heuristic value: " + minimalHValue);
			printMsg("");
			printMsg("------------------ END ------------------");
			printMsg("");
		}

		protected virtual bool shouldStop()
        {
			return stopWatch.Elapsed > timelimit;
		}

		protected virtual void Solve(TimeSpan maxTime)
        {
			printMsg("Search started");
			stopWatch = System.Diagnostics.Stopwatch.StartNew();

			expansions = 0;	minimalHValue = long.MaxValue; maxGValue = 0;
			totalExpansions = 0; totalBranching = 0; heuristicCalls = 0; totalHeuristicValue = 0;
			searchResult = default;
			this.timelimit = maxTime;

			openList.Insert(0, initialState);
            visited.Add(initialState, new StateValPair<S, P>(null, 0));

            while (openList.Size() > 0)
            {
                expansions++;
                if (expansions % 10000 == 0)
                {
					printProgress();
					if (shouldStop())
                    {
                        printMsg("Time limit exceeded.");
                        break;
                    }
                }

                var currentState = openList.RemoveMin();
                int currentGVal = visited[currentState].gValue;
				if (currentGVal > maxGValue)
					maxGValue = currentGVal;

				if (currentState.IsGoalState())
                {
					printResult(currentState);
					searchResult = currentState;
					return;
                }

				currentState.GetSuccessors(successorPlaceHolder);

                totalBranching += successorPlaceHolder.Count;
                totalExpansions++;

                foreach (var successor in successorPlaceHolder)
                {
					int successorGVal = currentGVal + successor.TransitionCost;
					var isUsefull = AddToClosedList(successor.State, currentState, successorGVal);
					if (!isUsefull)
						continue;

                    int hVal = h.GetValue(successor.State);
                    heuristicCalls++;
                    totalHeuristicValue += hVal;
                    if (hVal < minimalHValue)
                        minimalHValue = hVal;
					AddToOpenList(successor.State, successorGVal, hVal);
                }
            }
            Console.WriteLine("No solution found");
        }

        /// <summary>
        /// Returns the path begning with the initial state and ending in <paramref name="goalState"/>.
        /// Only to be called after the "search" finishes.
        /// </summary>
        /// <param name="goalState"></param>
        /// <returns></returns>
        protected List<S> extractSolution(S goalState)
        {
            List<S> solution = new List<S>();

            S currentState = goalState;
			solution.Add(goalState);
            while (currentState != default)
            {
                S pred = visited[currentState].predecessor;
                if (pred != default)
                    solution.Insert(0, pred);
                currentState = pred;
            }
            return solution;
        }
    }

	class GreedyBFS<S, P> : Astar<S, P> where S : State<S, P> where P : Problem<S, P>
    {
        protected override void AddToOpenList(S state, int gVal, int hVal)
        {
			openList.Insert(hVal, state);
		}
    }

	struct StateValPair<S, P> where P : Problem<S, P> where S : State<S, P>
	{
		public S predecessor;
		public int gValue;

		public StateValPair(S state, int val)
		{
			this.predecessor = state;
			this.gValue = val;
		}
	}


}
